from collections.abc import Sequence
from typing import TypedDict, Union, cast

from dagster._annotations import beta
from dagster._core.definitions.asset_spec import AssetSpec
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.definitions_class import Definitions

from dagster_airlift.constants import TASK_MAPPING_METADATA_KEY
from dagster_airlift.core import (
    build_defs_from_airflow_instance as build_defs_from_airflow_instance,
)
from dagster_airlift.core.load_defs import replace_assets_in_defs
from dagster_airlift.core.top_level_dag_def_api import spec_with_metadata


@beta
class TaskHandleDict(TypedDict):
    dag_id: str
    task_id: str


@beta
def assets_with_multiple_task_mappings(
    assets: Sequence[Union[AssetSpec, AssetsDefinition]], task_handles: list[TaskHandleDict]
) -> Sequence[Union[AssetSpec, AssetsDefinition]]:
    """Given an asset or assets definition, return a new asset or assets definition with metadata
    that indicates that it is targeted by multiple airflow tasks. An example of this would
    be a separate weekly and daily dag that contains a task that targets a single asset.

    .. code-block:: python

        from dagster import Definitions, AssetSpec, asset
        from dagster_airlift import (
            build_defs_from_airflow_instance,
            targeted_by_multiple_tasks,
            assets_with_task_mappings,
        )

        # Asset maps to a single task.
        @asset
        def other_asset(): ...

        # Asset maps to a physical entity which is produced by two different airflow tasks.
        @asset
        def scheduled_twice(): ...

        defs = build_defs_from_airflow_instance(
            airflow_instance=airflow_instance,
            defs=Definitions(
                assets=[
                    *assets_with_task_mappings(
                        dag_id="other_dag",
                        task_mappings={
                            "task1": [other_asset]
                        },
                    ),
                    *assets_with_multiple_task_mappings(
                        assets=[scheduled_twice],
                        task_handles=[
                            {"dag_id": "weekly_dag", "task_id": "task1"},
                            {"dag_id": "daily_dag", "task_id": "task1"},
                        ],
                    ),
                ]
            ),
        )

    """
    return [
        asset.map_asset_specs(
            lambda spec: spec_with_metadata(spec, {TASK_MAPPING_METADATA_KEY: task_handles})
        )
        if isinstance(asset, AssetsDefinition)
        else spec_with_metadata(asset, {TASK_MAPPING_METADATA_KEY: task_handles})
        for asset in assets
    ]


def targeted_by_multiple_tasks(
    defs: Definitions, task_handles: list[TaskHandleDict]
) -> Definitions:
    return replace_assets_in_defs(
        defs,
        assets_with_multiple_task_mappings(
            cast(Sequence[Union[AssetSpec, AssetsDefinition]], defs.assets),
            task_handles=task_handles,
        ),
    )



from collections.abc import Iterable, Mapping, Sequence
from typing import Any, Union

from dagster import (
    AssetsDefinition,
    AssetSpec,
    Definitions,
    _check as check,
)
from dagster._annotations import beta

from dagster_airlift.core.utils import metadata_for_dag_mapping, metadata_for_task_mapping


class TaskDefs:
    def __init__(self, task_id: str, defs: Definitions):
        self.task_id = task_id
        self.defs = defs


def apply_metadata_to_assets(
    assets: Iterable[Union[AssetsDefinition, AssetSpec]], metadata: dict[str, Any]
) -> Sequence[Union[AssetsDefinition, AssetSpec]]:
    return [assets_def_with_af_metadata(asset, metadata) for asset in assets]


def apply_metadata_to_all_specs(defs: Definitions, metadata: dict[str, Any]) -> Definitions:
    return Definitions(
        assets=[
            # Right now we make assumptions that we only support AssetSpec and AssetsDefinition
            # in orchesrated_defs.
            # https://linear.app/dagster-labs/issue/FOU-369/support-cacheableassetsdefinition-and-sourceasset-in-airlift
            assets_def_with_af_metadata(
                check.inst(
                    asset,
                    (AssetSpec, AssetsDefinition),
                    "Only supports AssetSpec and AssetsDefinition right now",
                ),
                metadata,
            )
            for asset in (defs.assets or [])
        ],
        resources=defs.resources,
        sensors=defs.sensors,
        schedules=defs.schedules,
        jobs=defs.jobs,
        loggers=defs.loggers,
        executor=defs.executor,
        asset_checks=defs.asset_checks,
    )


def spec_with_metadata(spec: AssetSpec, metadata: Mapping[str, Any]) -> "AssetSpec":
    return spec._replace(metadata={**spec.metadata, **metadata})


def assets_def_with_af_metadata(
    assets_def: Union[AssetsDefinition, AssetSpec], metadata: Mapping[str, str]
) -> Union[AssetsDefinition, AssetSpec]:
    return (
        assets_def.map_asset_specs(lambda spec: spec_with_metadata(spec, metadata))
        if isinstance(assets_def, AssetsDefinition)
        else spec_with_metadata(assets_def, metadata)
    )


@beta
def assets_with_task_mappings(
    dag_id: str, task_mappings: Mapping[str, Iterable[Union[AssetsDefinition, AssetSpec]]]
) -> Sequence[Union[AssetsDefinition, AssetSpec]]:
    """Modify assets to be associated with a particular task in Airlift tooling.

    Used in concert with `build_defs_from_airflow_instance` to observe an airflow
    instance to monitor the tasks that are associated with the assets and
    keep their materialization histories up to date.

    Concretely this adds metadata to all asset specs in the provided definitions
    with the provided dag_id and task_id. The dag_id comes from the dag_id argument;
    the task_id comes from the key of the provided task_mappings dictionary.
    There is a single metadata key "airlift/task-mapping" that is used to store
    this information. It is a list of dictionaries with keys "dag_id" and "task_id".

    Example:

        .. code-block:: python

            from dagster import AssetSpec, Definitions, asset
            from dagster_airlift.core import assets_with_task_mappings

            @asset
            def asset_one() -> None: ...

            defs = Definitions(
                assets=assets_with_task_mappings(
                    dag_id="dag_one",
                    task_mappings={
                        "task_one": [asset_one],
                        "task_two": [AssetSpec(key="asset_two"), AssetSpec(key="asset_three")],
                    },
                )
            )
    """
    assets_list = []
    for task_id, assets in task_mappings.items():
        assets_list.extend(
            apply_metadata_to_assets(
                assets,
                metadata_for_task_mapping(task_id=task_id, dag_id=dag_id),
            )
        )
    return assets_list


@beta
def assets_with_dag_mappings(
    dag_mappings: Mapping[str, Iterable[Union[AssetsDefinition, AssetSpec]]],
) -> Sequence[Union[AssetsDefinition, AssetSpec]]:
    """Modify assets to be associated with a particular dag in Airlift tooling.

    Used in concert with `build_defs_from_airflow_instance` to observe an airflow
    instance to monitor the dags that are associated with the assets and
    keep their materialization histories up to date.

    In contrast with `assets_with_task_mappings`, which maps assets on a per-task basis, this is used in concert with
    `proxying_to_dagster` dag-level mappings where an entire dag is migrated at once.

    Concretely this adds metadata to all asset specs in the provided definitions
    with the provided dag_id. The dag_id comes from the key of the provided dag_mappings dictionary.
    There is a single metadata key "airlift/dag-mapping" that is used to store
    this information. It is a list of strings, where each string is a dag_id which the asset is associated with.

    Example:

    .. code-block:: python

        from dagster import AssetSpec, Definitions, asset
        from dagster_airlift.core import assets_with_dag_mappings

        @asset
        def asset_one() -> None: ...

        defs = Definitions(
            assets=assets_with_dag_mappings(
                dag_mappings={
                    "dag_one": [asset_one],
                    "dag_two": [AssetSpec(key="asset_two"), AssetSpec(key="asset_three")],
                },
            )
        )
    """
    assets_list = []
    for dag_id, assets in dag_mappings.items():
        assets_list.extend(
            apply_metadata_to_assets(
                assets,
                metadata_for_dag_mapping(dag_id=dag_id),
            )
        )
    return assets_list


@beta
def dag_defs(dag_id: str, *defs: TaskDefs) -> Definitions:
    """Construct a Dagster :py:class:`Definitions` object with definitions
    associated with a particular Dag in Airflow that is being tracked by Airlift tooling.

    Concretely this adds metadata to all asset specs in the provided definitions
    with the provided dag_id and task_id. There is a single metadata key
    "airlift/task_mapping" that is used to store this information. It is a list of
    dictionaries with keys "dag_id" and "task_id".

    Used in concert with :py:func:`task_defs`.

    Example:
    .. code-block:: python
        defs = dag_defs(
            "dag_one",
            task_defs("task_one", Definitions(assets=[AssetSpec(key="asset_one"]))),
            task_defs("task_two", Definitions(assets=[AssetSpec(key="asset_two"), AssetSpec(key="asset_three")])),
        )
    """
    defs_to_merge = []
    for task_def in defs:
        defs_to_merge.append(
            apply_metadata_to_all_specs(
                defs=task_def.defs,
                metadata=metadata_for_task_mapping(task_id=task_def.task_id, dag_id=dag_id),
            )
        )
    return Definitions.merge(*defs_to_merge)


@beta
def task_defs(task_id, defs: Definitions) -> TaskDefs:
    """Associate a set of definitions with a particular task in Airflow that is being tracked
    by Airlift tooling.
    """
    return TaskDefs(task_id, defs)



from collections.abc import Iterable, Iterator
from typing import TYPE_CHECKING, Optional, Union

from dagster import (
    AssetsDefinition,
    AssetSpec,
    SourceAsset,
    _check as check,
)
from dagster._core.definitions.cacheable_assets import CacheableAssetsDefinition
from dagster._core.definitions.utils import VALID_NAME_REGEX
from dagster._core.errors import DagsterInvariantViolationError
from dagster._core.storage.tags import KIND_PREFIX

from dagster_airlift.constants import (
    AIRFLOW_SOURCE_METADATA_KEY_PREFIX,
    DAG_MAPPING_METADATA_KEY,
    PEERED_DAG_MAPPING_METADATA_KEY,
    TASK_MAPPING_METADATA_KEY,
)

if TYPE_CHECKING:
    from dagster_airlift.core.serialization.serialized_data import DagHandle, TaskHandle


def convert_to_valid_dagster_name(name: str) -> str:
    """Converts a name to a valid dagster name by replacing invalid characters with underscores. / is converted to a double underscore."""
    return "".join(c if VALID_NAME_REGEX.match(c) else "__" if c == "/" else "_" for c in name)


def airflow_kind_dict() -> dict:
    return {f"{KIND_PREFIX}airflow": ""}


def airlift_mapped_kind_dict() -> dict:
    return {f"{KIND_PREFIX}airliftmapped": ""}


def spec_iterator(
    assets: Optional[
        Iterable[Union[AssetsDefinition, AssetSpec, SourceAsset, CacheableAssetsDefinition]]
    ],
) -> Iterator[AssetSpec]:
    for asset in assets or []:
        if isinstance(asset, AssetsDefinition):
            yield from asset.specs
        elif isinstance(asset, AssetSpec):
            yield asset
        else:
            raise DagsterInvariantViolationError(
                "Expected orchestrated defs to all be AssetsDefinitions or AssetSpecs."
            )


def metadata_for_task_mapping(*, task_id: str, dag_id: str) -> dict:
    return {TASK_MAPPING_METADATA_KEY: [{"dag_id": dag_id, "task_id": task_id}]}


def metadata_for_dag_mapping(*, dag_id: str) -> dict:
    return {DAG_MAPPING_METADATA_KEY: [{"dag_id": dag_id}]}


def get_metadata_key(instance_name: str) -> str:
    return f"{AIRFLOW_SOURCE_METADATA_KEY_PREFIX}/{instance_name}"


def is_task_mapped_asset_spec(spec: AssetSpec) -> bool:
    return TASK_MAPPING_METADATA_KEY in spec.metadata


def is_dag_mapped_asset_spec(spec: AssetSpec) -> bool:
    return DAG_MAPPING_METADATA_KEY in spec.metadata


def is_peered_dag_asset_spec(spec: AssetSpec) -> bool:
    return PEERED_DAG_MAPPING_METADATA_KEY in spec.metadata


def task_handles_for_spec(spec: AssetSpec) -> set["TaskHandle"]:
    from dagster_airlift.core.serialization.serialized_data import TaskHandle

    check.param_invariant(is_task_mapped_asset_spec(spec), "spec", "Must be mapped spec")
    task_handles = []
    for task_handle_dict in spec.metadata[TASK_MAPPING_METADATA_KEY]:
        task_handles.append(
            TaskHandle(dag_id=task_handle_dict["dag_id"], task_id=task_handle_dict["task_id"])
        )
    return set(task_handles)


def dag_handles_for_spec(spec: AssetSpec) -> set["DagHandle"]:
    from dagster_airlift.core.serialization.serialized_data import DagHandle

    check.param_invariant(is_dag_mapped_asset_spec(spec), "spec", "Must be mapped spec")
    return {
        DagHandle(dag_id=dag_handle_dict["dag_id"])
        for dag_handle_dict in spec.metadata[DAG_MAPPING_METADATA_KEY]
    }


def peered_dag_handles_for_spec(spec: AssetSpec) -> set["DagHandle"]:
    from dagster_airlift.core.serialization.serialized_data import DagHandle

    check.param_invariant(is_peered_dag_asset_spec(spec), "spec", "Must be mapped spec")
    return {
        DagHandle(dag_id=dag_handle_dict["dag_id"])
        for dag_handle_dict in spec.metadata[PEERED_DAG_MAPPING_METADATA_KEY]
    }



from collections import defaultdict
from collections.abc import Sequence
from datetime import datetime, timedelta
from typing import Any, Optional

import requests

from dagster_airlift.core import AirflowInstance
from dagster_airlift.core.airflow_instance import DagInfo, DagRun, TaskInfo, TaskInstance
from dagster_airlift.core.basic_auth import AirflowAuthBackend


class DummyAuthBackend(AirflowAuthBackend):
    def get_session(self) -> requests.Session:
        raise NotImplementedError("This shouldn't be called from this mock context.")

    def get_webserver_url(self) -> str:
        return "http://dummy.domain"


class AirflowInstanceFake(AirflowInstance):
    """Loads a set of provided DagInfo, TaskInfo, and TaskInstance objects for testing."""

    def __init__(
        self,
        dag_infos: list[DagInfo],
        task_infos: list[TaskInfo],
        task_instances: list[TaskInstance],
        dag_runs: list[DagRun],
        variables: list[dict[str, Any]] = [],
        instance_name: Optional[str] = None,
        max_runs_per_batch: Optional[int] = None,
    ) -> None:
        self._dag_infos_by_dag_id = {dag_info.dag_id: dag_info for dag_info in dag_infos}
        self._task_infos_by_dag_and_task_id = {
            (task_info.dag_id, task_info.task_id): task_info for task_info in task_infos
        }
        self._task_instances_by_dag_and_task_id: dict[tuple[str, str], list[TaskInstance]] = (
            defaultdict(list)
        )
        for task_instance in task_instances:
            self._task_instances_by_dag_and_task_id[
                (task_instance.dag_id, task_instance.task_id)
            ].append(task_instance)
        self._dag_runs_by_dag_id: dict[str, list[DagRun]] = defaultdict(list)
        for dag_run in dag_runs:
            self._dag_runs_by_dag_id[dag_run.dag_id].append(dag_run)
        self._dag_infos_by_file_token = {dag_info.file_token: dag_info for dag_info in dag_infos}
        self._variables = variables
        self._max_runs_per_batch = max_runs_per_batch
        super().__init__(
            auth_backend=DummyAuthBackend(),
            name="test_instance" if instance_name is None else instance_name,
        )

    def list_dags(self) -> list[DagInfo]:
        return list(self._dag_infos_by_dag_id.values())

    def list_variables(self) -> list[dict[str, Any]]:
        return self._variables

    def get_dag_runs(self, dag_id: str, start_date: datetime, end_date: datetime) -> list[DagRun]:
        if dag_id not in self._dag_runs_by_dag_id:
            raise ValueError(f"Dag run not found for dag_id {dag_id}")
        return [
            run
            for run in self._dag_runs_by_dag_id[dag_id]
            if start_date.timestamp() <= run.start_date.timestamp() <= end_date.timestamp()
            and start_date.timestamp() <= run.end_date.timestamp() <= end_date.timestamp()
        ]

    def get_dag_runs_batch(
        self,
        dag_ids: Sequence[str],
        end_date_gte: datetime,
        end_date_lte: datetime,
        offset: int = 0,
    ) -> tuple[list[DagRun], int]:
        runs = [
            (run.end_date, run)
            for runs in self._dag_runs_by_dag_id.values()
            for run in runs
            if end_date_gte.timestamp() <= run.end_date.timestamp() <= end_date_lte.timestamp()
            and run.dag_id in dag_ids
        ]
        sorted_by_end_date = [run for _, run in sorted(runs, key=lambda x: x[0])]
        end_idx = (
            offset + self._max_runs_per_batch
            if self._max_runs_per_batch
            else len(sorted_by_end_date)
        )
        return (sorted_by_end_date[offset:end_idx], len(sorted_by_end_date))

    def get_task_instance_batch(
        self, dag_id: str, task_ids: Sequence[str], run_id: str, states: Sequence[str]
    ) -> list[TaskInstance]:
        task_instances = []
        for task_id in set(task_ids):
            if (dag_id, task_id) not in self._task_instances_by_dag_and_task_id:
                continue
            task_instances.extend(
                [
                    task_instance
                    for task_instance in self._task_instances_by_dag_and_task_id[(dag_id, task_id)]
                    if task_instance.run_id == run_id and task_instance.state in states
                ]
            )
        return task_instances

    def get_task_instance(self, dag_id: str, task_id: str, run_id: str) -> TaskInstance:
        if (dag_id, task_id) not in self._task_instances_by_dag_and_task_id:
            raise ValueError(f"Task instance not found for dag_id {dag_id} and task_id {task_id}")
        if not any(
            task_instance.run_id == run_id
            for task_instance in self._task_instances_by_dag_and_task_id[(dag_id, task_id)]
        ):
            raise ValueError(
                f"Task instance not found for dag_id {dag_id}, task_id {task_id}, and run_id {run_id}"
            )
        return next(
            iter(
                task_instance
                for task_instance in self._task_instances_by_dag_and_task_id[(dag_id, task_id)]
                if task_instance.run_id == run_id
            )
        )

    def get_task_info(self, dag_id, task_id) -> TaskInfo:
        if (dag_id, task_id) not in self._task_infos_by_dag_and_task_id:
            raise ValueError(f"Task info not found for dag_id {dag_id} and task_id {task_id}")
        return self._task_infos_by_dag_and_task_id[(dag_id, task_id)]

    def get_task_infos(self, *, dag_id: str) -> list[TaskInfo]:
        if dag_id not in self._dag_infos_by_dag_id:
            raise ValueError(f"Dag info not found for dag_id {dag_id}")
        task_infos = []
        for dag_id, task_id in self._task_infos_by_dag_and_task_id:
            if dag_id == dag_id:
                task_infos.append(self._task_infos_by_dag_and_task_id[(dag_id, task_id)])
        return task_infos

    def get_dag_info(self, dag_id) -> DagInfo:
        if dag_id not in self._dag_infos_by_dag_id:
            raise ValueError(f"Dag info not found for dag_id {dag_id}")
        return self._dag_infos_by_dag_id[dag_id]

    def get_dag_source_code(self, file_token: str) -> str:
        if file_token not in self._dag_infos_by_file_token:
            raise ValueError(f"Dag info not found for file_token {file_token}")
        return "indicates found source code"


def make_dag_info(dag_id: str, file_token: Optional[str]) -> DagInfo:
    return DagInfo(
        webserver_url="http://dummy.domain",
        dag_id=dag_id,
        metadata={"file_token": file_token if file_token else "dummy_file_token"},
    )


def make_task_info(
    dag_id: str, task_id: str, downstream_task_ids: Optional[list[str]] = None
) -> TaskInfo:
    return TaskInfo(
        webserver_url="http://dummy.domain",
        dag_id=dag_id,
        task_id=task_id,
        metadata={"downstream_task_ids": downstream_task_ids or []},
    )


def make_task_instance(
    dag_id: str,
    task_id: str,
    run_id: str,
    start_date: datetime,
    end_date: datetime,
    logical_date: Optional[datetime] = None,
) -> TaskInstance:
    return TaskInstance(
        webserver_url="http://dummy.domain",
        dag_id=dag_id,
        task_id=task_id,
        run_id=run_id,
        metadata={
            "state": "success",
            "start_date": start_date.isoformat(),
            "end_date": end_date.isoformat(),
            "logical_date": logical_date.isoformat() if logical_date else start_date.isoformat(),
        },
    )


def make_dag_run(
    dag_id: str,
    run_id: str,
    start_date: datetime,
    end_date: datetime,
    logical_date: Optional[datetime] = None,
) -> DagRun:
    return DagRun(
        webserver_url="http://dummy.domain",
        dag_id=dag_id,
        run_id=run_id,
        metadata={
            "state": "success",
            "start_date": start_date.isoformat(),
            "end_date": end_date.isoformat(),
            "logical_date": logical_date.isoformat() if logical_date else start_date.isoformat(),
            "run_type": "manual",
            "note": "dummy note",
            "conf": {},
        },
    )


def make_instance(
    dag_and_task_structure: dict[str, list[str]],
    dag_runs: list[DagRun] = [],
    task_deps: dict[str, list[str]] = {},
    instance_name: Optional[str] = None,
    max_runs_per_batch: Optional[int] = None,
) -> AirflowInstanceFake:
    """Constructs DagInfo, TaskInfo, and TaskInstance objects from provided data.

    Args:
        dag_and_task_structure: A dictionary mapping dag_id to a list of task_ids.
        dag_runs: A list of DagRun objects to include in the instance. A TaskInstance object will be
            created for each task_id in the dag, for each DagRun in dag_runs pertaining to a particular dag.
    """
    dag_infos = []
    task_infos = []
    for dag_id, task_ids in dag_and_task_structure.items():
        dag_info = make_dag_info(dag_id=dag_id, file_token=dag_id)
        dag_infos.append(dag_info)
        task_infos.extend(
            [
                make_task_info(
                    dag_id=dag_id, task_id=task_id, downstream_task_ids=task_deps.get(task_id, [])
                )
                for task_id in task_ids
            ]
        )
    task_instances = []
    for dag_run in dag_runs:
        task_instances.extend(
            [
                make_task_instance(
                    dag_id=dag_run.dag_id,
                    task_id=task_id,
                    run_id=dag_run.run_id,
                    start_date=dag_run.start_date,
                    end_date=dag_run.end_date
                    - timedelta(
                        seconds=1
                    ),  # Ensure that the task ends before the full "dag" completes.
                    logical_date=dag_run.logical_date,
                )
                for task_id in dag_and_task_structure[dag_run.dag_id]
            ]
        )
    return AirflowInstanceFake(
        dag_infos=dag_infos,
        task_infos=task_infos,
        task_instances=task_instances,
        dag_runs=dag_runs,
        instance_name=instance_name,
        max_runs_per_batch=max_runs_per_batch,
    )



import subprocess
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path

from dagster._core.test_utils import environ


def airlift_root() -> Path:
    return Path(__file__).parent.parent.parent


def remove_airflow_home_remnants(airflow_home: Path) -> None:
    logs_path = airflow_home / "logs"
    cfg_path = airflow_home / "airflow.cfg"
    db_path = airflow_home / "airflow.db"
    proxied_state_path = airflow_home / "airflow_dags" / "proxied_state"

    subprocess.check_output(
        ["rm", "-rf", str(logs_path), str(cfg_path), str(db_path), str(proxied_state_path)]
    )


def airflow_cfg_script_path() -> Path:
    return airlift_root() / "scripts" / "airflow_setup.sh"


def chmod_script(script_path: Path) -> None:
    subprocess.check_output(["chmod", "+x", str(script_path)])


@contextmanager
def configured_airflow_home(airflow_home: Path) -> Generator[None, None, None]:
    path_to_dags = airflow_home / "airflow_dags"
    with environ({"AIRFLOW_HOME": str(airflow_home)}):
        try:
            # Start by removing old cruft if it exists
            remove_airflow_home_remnants(airflow_home)
            # Scaffold the airflow configuration file.
            chmod_script(airflow_cfg_script_path())
            subprocess.run(
                [str(airflow_cfg_script_path()), str(path_to_dags), str(airflow_home)], check=False
            )
            yield
        finally:
            # Clean up after ourselves.
            remove_airflow_home_remnants(airflow_home)



from importlib import util as iutil

from dagster_airlift.version import __version__

is_dagster_installed = iutil.find_spec("dagster") is not None
# In the case of running CI/CD, where we have Dagster installed in the environment, we want to be able to register the library in our library registry. But we want
# to avoid taking an actual dependency on Dagster, so we need to gate this behind a check for the presence of the Dagster package in the environment.
if is_dagster_installed:
    from dagster._core.libraries import DagsterLibraryRegistry

    DagsterLibraryRegistry.register("dagster-airlift", __version__)



from dagster_airlift.core.airflow_defs_data import AirflowDefinitionsData as AirflowDefinitionsData
from dagster_airlift.core.basic_auth import (
    AirflowAuthBackend as AirflowAuthBackend,
    AirflowBasicAuthBackend as AirflowBasicAuthBackend,
)
from dagster_airlift.core.load_defs import (
    AirflowInstance as AirflowInstance,
    DagSelectorFn as DagSelectorFn,
    build_defs_from_airflow_instance as build_defs_from_airflow_instance,
    load_airflow_dag_asset_specs as load_airflow_dag_asset_specs,
)
from dagster_airlift.core.multiple_tasks import (
    TaskHandleDict as TaskHandleDict,
    assets_with_multiple_task_mappings as assets_with_multiple_task_mappings,
)
from dagster_airlift.core.sensor.event_translation import (
    AssetEvent as AssetEvent,
    DagsterEventTransformerFn as DagsterEventTransformerFn,
)
from dagster_airlift.core.sensor.sensor_builder import (
    build_airflow_polling_sensor as build_airflow_polling_sensor,
)
from dagster_airlift.core.serialization.serialized_data import DagInfo as DagInfo
from dagster_airlift.core.top_level_dag_def_api import (
    assets_with_dag_mappings as assets_with_dag_mappings,
    assets_with_task_mappings as assets_with_task_mappings,
    dag_defs as dag_defs,
    task_defs as task_defs,
)



from collections.abc import Sequence
from pathlib import Path
from typing import Any, NamedTuple, Optional

import yaml
from dagster._annotations import beta


@beta
class TaskProxiedState(NamedTuple):
    """A class to store the proxied state of a task.

    Args:
        task_id (str): The id of the task.
        proxied (bool): A boolean indicating whether the task is proxied.
    """

    task_id: str
    proxied: bool

    @staticmethod
    def from_dict(task_dict: dict[str, Any]) -> "TaskProxiedState":
        if set(task_dict.keys()) != {"id", "proxied"}:
            raise Exception(
                f"Expected 'proxied' and 'id' keys in the task dictionary. Found keys: {task_dict.keys()}"
            )
        if task_dict["proxied"] not in [True, False]:
            raise Exception("Expected 'proxied' key to be a boolean")
        return TaskProxiedState(task_id=task_dict["id"], proxied=task_dict["proxied"])

    def to_dict(self) -> dict[str, Any]:
        return {"id": self.task_id, "proxied": self.proxied}


@beta
class DagProxiedState(NamedTuple):
    """A class to store the proxied state of tasks in a dag.

    Args:
        tasks (Dict[str, TaskProxiedState]): A dictionary of task_id to TaskProxiedState. If the entire dag is proxied, or proxied state
            is not set for a task, the task_id will not be present in this dictionary.
        proxied (Optional[bool]): A boolean indicating whether the entire dag is proxied. If this is None, then the dag proxies at the task level (or
        proxying state has not been set at all).
    """

    proxied: Optional[bool]
    tasks: dict[str, TaskProxiedState]

    @staticmethod
    def from_dict(dag_dict: dict[str, Any]) -> "DagProxiedState":
        if "tasks" not in dag_dict and "proxied" not in dag_dict:
            raise Exception(
                f"Expected a 'tasks' or 'proxied' top-level key in the dag dictionary. Instead; got: {dag_dict}"
            )
        if "tasks" in dag_dict and "proxied" in dag_dict:
            raise Exception(
                f"Expected only one of 'tasks' or 'proxied' top-level keys in the dag dictionary. Instead; got: {dag_dict}"
            )
        task_proxied_states = {}
        if "tasks" in dag_dict:
            task_list: Sequence[dict[str, Any]] = dag_dict["tasks"]
            for task_dict in task_list:
                task_state = TaskProxiedState.from_dict(task_dict)
                task_proxied_states[task_state.task_id] = task_state
        dag_proxied_state: Optional[bool] = dag_dict.get("proxied")
        if dag_proxied_state not in [True, False, None]:
            raise Exception("Expected 'proxied' key to be a boolean or None")
        return DagProxiedState(tasks=task_proxied_states, proxied=dag_proxied_state)

    def to_dict(self) -> dict[str, Sequence[dict[str, Any]]]:
        return {"tasks": [task_state.to_dict() for task_state in self.tasks.values()]}

    def is_task_proxied(self, task_id: str) -> bool:
        if task_id not in self.tasks:
            return False
        return self.tasks[task_id].proxied

    @property
    def dag_proxies_at_task_level(self) -> bool:
        """Dags can proxy on either a task-by-task basis, or for the entire dag at once.
        We use the proxied state to determine which is the case for a given dag. If the dag's proxied state
        is None, then we assume the dag proxies at the task level. If the dag's proxied state is a boolean,
        then we assume the dag proxies at the dag level.
        """
        return self.proxied is None

    @property
    def dag_proxies_at_dag_level(self) -> bool:
        """Dags can proxy on either a task-by-task basis, or for the entire dag at once.
        We use the proxied state to determine which is the case for a given dag. If the dag's proxied state
        is None, then we assume the dag proxies at the task level. If the dag's proxied state is a boolean,
        then we assume the dag proxies at the dag level.
        """
        return self.proxied is not None


@beta
class AirflowProxiedState(NamedTuple):
    """A class to store the proxied state of dags and tasks in Airflow.
    Typically, this is constructed by :py:func:`load_proxied_state_from_yaml`.

    Args:
        dags (Dict[str, DagProxiedState]): A dictionary of dag_id to DagProxiedState.
    """

    dags: dict[str, DagProxiedState]

    def get_task_proxied_state(self, *, dag_id: str, task_id: str) -> Optional[bool]:
        if dag_id not in self.dags:
            return None
        if task_id not in self.dags[dag_id].tasks:
            return None
        return self.dags[dag_id].tasks[task_id].proxied

    def dag_has_proxied_state(self, dag_id: str) -> bool:
        return self.get_proxied_dict_for_dag(dag_id) is not None

    def dag_proxies_at_task_level(self, dag_id: str) -> bool:
        """Dags can proxy on either a task-by-task basis, or for the entire dag at once.
        We use the proxied state to determine which is the case for a given dag. If the dag's proxied state
        is None, then we assume the dag proxies at the task level. If the dag's proxied state is a boolean,
        then we assume the dag proxies at the dag level.
        """
        return self.dags[dag_id].dag_proxies_at_task_level

    def dag_proxies_at_dag_level(self, dag_id: str) -> bool:
        """Dags can proxy on either a task-by-task basis, or for the entire dag at once.
        We use the proxied state to determine which is the case for a given dag. If the dag's proxied state
        is None, then we assume the dag proxies at the task level. If the dag's proxied state is a boolean,
        then we assume the dag proxies at the dag level.
        """
        return self.dags[dag_id].dag_proxies_at_dag_level

    def get_proxied_dict_for_dag(
        self, dag_id: str
    ) -> Optional[dict[str, Sequence[dict[str, Any]]]]:
        if dag_id not in self.dags:
            return None
        return {
            "tasks": [
                {"proxied": task_state.proxied, "id": task_id}
                for task_id, task_state in self.dags[dag_id].tasks.items()
            ]
        }

    @staticmethod
    def from_dict(proxied_dict: dict[str, Any]) -> "AirflowProxiedState":
        dags = {}
        for dag_id, dag_dict in proxied_dict.items():
            dags[dag_id] = DagProxiedState.from_dict(dag_dict)
        return AirflowProxiedState(dags=dags)


class ProxiedStateParsingError(Exception):
    pass


@beta
def load_proxied_state_from_yaml(proxied_yaml_path: Path) -> AirflowProxiedState:
    """Loads the proxied state from a directory of yaml files.

    Expects the directory to contain yaml files, where each file corresponds to the id of a dag (ie: `dag_id.yaml`).
    This directory is typically constructed using the `dagster-airlift` CLI:

        .. code-block:: bash

            AIRFLOW_HOME=... dagster-airlift proxy scaffold

    The file should have either of the following structure.
    In the case of task-level proxying:

        .. code-block:: yaml

                tasks:
                    - id: task_id
                      proxied: true
                    - id: task_id
                      proxied: false

    In the case of dag-level proxying:

        .. code-block:: yaml

                proxied: true

    Args:
        proxied_yaml_path (Path): The path to the directory containing the yaml files.

    Returns:
        AirflowProxiedState: The proxied state of the dags and tasks in Airflow.
    """
    # Expect proxied_yaml_path to be a directory, where each file represents a dag, and each
    # file in the subdir represents a task. The dictionary for each task should contain two keys;
    # id: the task id, and proxied: a boolean indicating whether the task has been proxied.
    dag_proxied_states = {}
    try:
        for dag_file in proxied_yaml_path.iterdir():
            # Check that the file is a yaml file or yml file
            if dag_file.suffix not in [".yaml", ".yml"]:
                continue
            dag_id = dag_file.stem
            yaml_dict = yaml.safe_load(dag_file.read_text())
            if not isinstance(yaml_dict, dict):
                raise Exception("Expected a dictionary")
            dag_proxied_states[dag_id] = DagProxiedState.from_dict(yaml_dict)
    except Exception as e:
        raise ProxiedStateParsingError("Error parsing proxied state yaml") from e
    return AirflowProxiedState(dags=dag_proxied_states)



from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from typing import AbstractSet, Any, Callable, Union, cast  # noqa: UP035

from dagster import (
    AssetMaterialization,
    AssetObservation,
    JsonMetadataValue,
    MarkdownMetadataValue,
    SensorEvaluationContext,
    TimestampMetadataValue,
    _check as check,
)
from dagster._core.definitions.asset_check_evaluation import AssetCheckEvaluation
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.time_window_partitions import TimeWindowPartitionsDefinition
from dagster._time import datetime_from_timestamp, get_current_timestamp

from dagster_airlift.constants import (
    AIRFLOW_RUN_ID_METADATA_KEY,
    AIRFLOW_TASK_INSTANCE_LOGICAL_DATE_METADATA_KEY,
    EFFECTIVE_TIMESTAMP_METADATA_KEY,
)
from dagster_airlift.core.airflow_defs_data import AirflowDefinitionsData
from dagster_airlift.core.airflow_instance import DagRun, TaskInstance
from dagster_airlift.core.serialization.serialized_data import DagHandle

AssetEvent = Union[AssetMaterialization, AssetObservation, AssetCheckEvaluation]
DagsterEventTransformerFn = Callable[
    [SensorEvaluationContext, AirflowDefinitionsData, Sequence[AssetMaterialization]],
    Iterable[AssetEvent],
]


def default_event_transformer(
    context: SensorEvaluationContext,
    airflow_data: AirflowDefinitionsData,
    materializations: Sequence[AssetMaterialization],
) -> Iterable[AssetEvent]:
    """The default event transformer function, which attaches a partition key to materializations which are from time-window partitioned assets."""
    cached_partition_calculations = defaultdict(dict)
    for mat in materializations:
        asset_spec = airflow_data.all_asset_specs_by_key[mat.asset_key]
        if not asset_spec.partitions_def or not isinstance(
            asset_spec.partitions_def, TimeWindowPartitionsDefinition
        ):
            yield mat
            continue
        airflow_logical_date_timestamp: float = cast(
            TimestampMetadataValue, mat.metadata[AIRFLOW_TASK_INSTANCE_LOGICAL_DATE_METADATA_KEY]
        ).value
        partitions_def = cast(TimeWindowPartitionsDefinition, asset_spec.partitions_def)
        calcs_for_def = cached_partition_calculations[partitions_def]
        if airflow_logical_date_timestamp not in calcs_for_def:
            cached_partition_calculations[partitions_def][airflow_logical_date_timestamp] = (
                get_partition_key_from_timestamp(
                    partitions_def=cast(TimeWindowPartitionsDefinition, asset_spec.partitions_def),
                    timestamp=airflow_logical_date_timestamp,
                )
            )
        partition = cached_partition_calculations[partitions_def][airflow_logical_date_timestamp]
        partitioned_mat = mat._replace(partition=partition)
        yield partitioned_mat


def get_partition_key_from_timestamp(
    partitions_def: TimeWindowPartitionsDefinition,
    timestamp: float,
) -> str:
    datetime_in_tz = datetime_from_timestamp(timestamp, partitions_def.timezone)
    # Assuming that "logical_date" lies on a partition, the previous partition window
    # (where upper bound can be the passed-in date, which is why we set respect_bounds=False)
    # will end on the logical date. This would indicate that there is a partition for the logical date.
    partition_window = check.not_none(
        partitions_def.get_prev_partition_window(datetime_in_tz, respect_bounds=False),
        f"Could not find partition for airflow logical date {datetime_in_tz.isoformat()}. This likely means that your partition range is too small to cover the logical date.",
    )
    check.invariant(
        datetime_in_tz.timestamp() == partition_window.end.timestamp(),
        (
            f"Expected logical date {datetime_in_tz.isoformat()} to match a partition in the partitions definition. This likely means that "
            "The partition range is not aligned with the scheduling interval in airflow."
        ),
    )
    check.invariant(
        datetime_in_tz.timestamp() >= partitions_def.start.timestamp(),
        (
            "provided date is before the start of the partitions definition. "
            "Ensure that the start date of your PartitionsDefinition is early enough to capture the provided date {datetime_in_tz.isoformat()}."
        ),
    )
    return partitions_def.get_partition_key_for_timestamp(timestamp)


def get_timestamp_from_materialization(event: AssetEvent) -> float:
    return check.float_param(
        event.metadata[EFFECTIVE_TIMESTAMP_METADATA_KEY].value,
        "Materialization Effective Timestamp",
    )


def synthetic_mats_for_peered_dag_asset_keys(
    dag_run: DagRun, airflow_data: AirflowDefinitionsData
) -> Sequence[AssetMaterialization]:
    return [
        dag_synthetic_mat(dag_run, airflow_data, asset_key)
        for asset_key in airflow_data.peered_dag_asset_keys_by_dag_handle[DagHandle(dag_run.dag_id)]
    ]


def synthetic_mats_for_mapped_dag_asset_keys(
    dag_run: DagRun, airflow_data: AirflowDefinitionsData
) -> Sequence[AssetMaterialization]:
    return [
        dag_synthetic_mat(dag_run, airflow_data, asset_key)
        for asset_key in airflow_data.mapped_asset_keys_by_dag_handle[DagHandle(dag_run.dag_id)]
    ]


def dag_synthetic_mat(
    dag_run: DagRun, airflow_data: AirflowDefinitionsData, asset_key: AssetKey
) -> AssetMaterialization:
    return AssetMaterialization(
        asset_key=asset_key, description=dag_run.note, metadata=get_dag_run_metadata(dag_run)
    )


def get_dag_run_metadata(dag_run: DagRun) -> Mapping[str, Any]:
    return {
        **get_common_metadata(dag_run),
        "Run Details": MarkdownMetadataValue(f"[View Run]({dag_run.url})"),
        "Start Date": TimestampMetadataValue(dag_run.start_date.timestamp()),
        "End Date": TimestampMetadataValue(dag_run.end_date.timestamp()),
        EFFECTIVE_TIMESTAMP_METADATA_KEY: TimestampMetadataValue(dag_run.end_date.timestamp()),
    }


def get_common_metadata(dag_run: DagRun) -> Mapping[str, Any]:
    return {
        "Airflow Run ID": dag_run.run_id,
        AIRFLOW_RUN_ID_METADATA_KEY: dag_run.run_id,
        "Run Metadata (raw)": JsonMetadataValue(dag_run.metadata),
        "Run Type": dag_run.run_type,
        "Airflow Config": JsonMetadataValue(dag_run.config),
        "Creation Timestamp": TimestampMetadataValue(get_current_timestamp()),
    }


def get_task_instance_metadata(dag_run: DagRun, task_instance: TaskInstance) -> Mapping[str, Any]:
    return {
        **get_common_metadata(dag_run),
        "Run Details": MarkdownMetadataValue(f"[View Run]({task_instance.details_url})"),
        "Task Logs": MarkdownMetadataValue(f"[View Logs]({task_instance.log_url})"),
        "Start Date": TimestampMetadataValue(task_instance.start_date.timestamp()),
        "End Date": TimestampMetadataValue(task_instance.end_date.timestamp()),
        EFFECTIVE_TIMESTAMP_METADATA_KEY: TimestampMetadataValue(
            task_instance.end_date.timestamp()
        ),
        AIRFLOW_TASK_INSTANCE_LOGICAL_DATE_METADATA_KEY: TimestampMetadataValue(
            task_instance.logical_date.timestamp()
        ),
    }


def synthetic_mats_for_task_instance(
    airflow_data: AirflowDefinitionsData,
    dag_run: DagRun,
    task_instance: TaskInstance,
) -> Sequence[AssetMaterialization]:
    asset_keys = airflow_data.asset_keys_in_task(dag_run.dag_id, task_instance.task_id)
    return synthetic_mats_for_mapped_asset_keys(dag_run, task_instance, asset_keys)


def synthetic_mats_for_mapped_asset_keys(
    dag_run: DagRun, task_instance: TaskInstance, asset_keys: AbstractSet[AssetKey]
) -> Sequence[AssetMaterialization]:
    mats = []
    for asset_key in asset_keys:
        mats.append(
            AssetMaterialization(
                asset_key=asset_key,
                description=task_instance.note,
                metadata=get_task_instance_metadata(dag_run, task_instance),
            )
        )
    return mats



# ruff: noqa: TID252
from .airflow_test_instance import (
    AirflowInstanceFake as AirflowInstanceFake,
    DummyAuthBackend as DummyAuthBackend,
    make_dag_info as make_dag_info,
    make_dag_run as make_dag_run,
    make_instance as make_instance,
    make_task_info as make_task_info,
    make_task_instance as make_task_instance,
)
from .test_utils import (
    airlift_root as airlift_root,
    configured_airflow_home as configured_airflow_home,
    remove_airflow_home_remnants as remove_airflow_home_remnants,
)



from collections.abc import Mapping
from functools import cached_property
from typing import AbstractSet, Any, NamedTuple, Optional  # noqa: UP035

from dagster import (
    AssetKey,
    _check as check,
)
from dagster._annotations import PublicAttr, beta
from dagster._record import record
from dagster._serdes import whitelist_for_serdes


@whitelist_for_serdes
@record
class TaskInfo:
    webserver_url: str
    dag_id: str
    task_id: str
    metadata: dict[str, Any]

    @property
    def dag_url(self) -> str:
        return f"{self.webserver_url}/dags/{self.dag_id}"

    @cached_property
    def downstream_task_ids(self) -> list[str]:
        return check.is_list(self.metadata["downstream_task_ids"], str)


@beta
@whitelist_for_serdes
@record
class DagInfo:
    """A record containing information about a given airflow dag.

    Users should not instantiate this class directly. It is provided when customizing which DAGs are included
    in the generated definitions using the `dag_selector_fn` argument of :py:func:`build_defs_from_airflow_instance`.

    Args:
        metadata (Dict[str, Any]): The metadata associated with the dag, retrieved by the Airflow REST API:
            https://airflow.apache.org/docs/apache-airflow/stable/stable-rest-api-ref.html#operation/get_dags
    """

    webserver_url: str
    dag_id: str
    metadata: PublicAttr[dict[str, Any]]

    @property
    def url(self) -> str:
        return f"{self.webserver_url}/dags/{self.dag_id}"

    @property
    def file_token(self) -> str:
        return self.metadata["file_token"]


@whitelist_for_serdes
class TaskHandle(NamedTuple):
    dag_id: str
    task_id: str


@whitelist_for_serdes
class DagHandle(NamedTuple):
    dag_id: str


###################################################################################################
# Serialized data that scopes to airflow DAGs and tasks.
###################################################################################################
# History:
# - created
@whitelist_for_serdes
@record
class SerializedDagData:
    """A record containing pre-computed data about a given airflow dag."""

    dag_id: str
    dag_info: DagInfo
    source_code: Optional[str]
    leaf_asset_keys: set[AssetKey]
    task_infos: Mapping[str, TaskInfo]


@whitelist_for_serdes
@record
class KeyScopedTaskHandles:
    asset_key: AssetKey
    mapped_tasks: AbstractSet[TaskHandle]


@whitelist_for_serdes
@record
class KeyScopedDagHandles:
    asset_key: AssetKey
    mapped_dags: AbstractSet[DagHandle]


###################################################################################################
# Serializable data that will be cached to avoid repeated calls to the Airflow API, and to avoid
# repeated scans of passed-in Definitions objects.
###################################################################################################
# History:
# - created
# - removed existing_asset_data
# - added key_scope_data_items
# - added instance_name
@whitelist_for_serdes
@record
class SerializedAirflowDefinitionsData:
    instance_name: str
    key_scoped_task_handles: list[KeyScopedTaskHandles]
    key_scoped_dag_handles: list[KeyScopedDagHandles]
    dag_datas: Mapping[str, SerializedDagData]

    @cached_property
    def all_mapped_tasks(self) -> dict[AssetKey, AbstractSet[TaskHandle]]:
        return {item.asset_key: item.mapped_tasks for item in self.key_scoped_task_handles}

    @cached_property
    def all_mapped_dags(self) -> dict[AssetKey, AbstractSet[DagHandle]]:
        return {item.asset_key: item.mapped_dags for item in self.key_scoped_dag_handles}



from collections.abc import Mapping, Sequence
from typing import Any, Optional

from dagster import AssetExecutionContext, AssetSpec, Definitions, multi_asset
from dagster._annotations import beta
from dagster_k8s import PipesK8sClient


@beta
def k8s_pod_defs(
    specs: Sequence[AssetSpec],
    pod_spec: Mapping[str, Any],
    pod_metadata: Mapping[str, Any],
    load_incluster_config: bool = False,
    kube_context: Optional[str] = None,
    kubeconfig_file: Optional[str] = None,
) -> Definitions:
    """Define a set of assets which are materialized in a k8s pod.

    This is designed to be an equivalent to the KubernetesPodOperator in Airflow.
    The same non-airflow specific arguments are exposed here as in the operator, and their meaning is maintained.
    Under the hood, this is launching an ephemeral kubernetes pod using the :py:class:`dagster_k8s.PipesK8sClient`.

    Comparison of Airflow KubernetesPodOperator parameters to this function.

    Directly supported arguments:
    - in_cluster (named load_incluster_config here)
    - cluster_context (named kube_context here)
    - config_file (named kubeconfig_file here)

    Many arguments are supported indirectly via the pod_spec argument.
    - volumes: Volumes to be used by the Pod (key ['volumes'])
    - affinity: Node affinity/anti-affinity rules for the Pod (key ['affinity'])
    - node_selector: Node selection constraints for the Pod (key ['nodeSelector'])
    - hostnetwork: Enable host networking for the Pod (key ['hostNetwork'])
    - dns_config: DNS settings for the Pod (key ['dnsConfig'])
    - dnspolicy: DNS policy for the Pod (key ['dnsPolicy'])
    - hostname: Hostname of the Pod (key ['hostname'])
    - subdomain: Subdomain for the Pod (key ['subdomain'])
    - schedulername: Scheduler to be used for the Pod (key ['schedulerName'])
    - service_account_name: Service account to be used by the Pod (key ['serviceAccountName'])
    - priority_class_name: Priority class for the Pod (key ['priorityClassName'])
    - security_context: Security context for the entire Pod (key ['securityContext'])
    - tolerations: Tolerations for the Pod (key ['tolerations'])
    - image_pull_secrets: Secrets for pulling container images (key ['imagePullSecrets'])
    - termination_grace_period: Grace period for Pod termination (key ['terminationGracePeriodSeconds'])
    - active_deadline_seconds: Deadline for the Pod's execution (key ['activeDeadlineSeconds'])
    - host_aliases: Additional entries for the Pod's /etc/hosts (key ['hostAliases'])
    - init_containers: Initialization containers for the Pod (key ['initContainers'])
    The following arguments are supported under the nested ['containers'] key of the pod_spec argument:
    - image: Docker image for the container (key ['image'])
    - cmds: Entrypoint command for the container (key ['command'])
    - arguments: Arguments for the entrypoint command (key ['args'])
    - ports: List of ports to expose from the container (key ['ports'])
    - volume_mounts: List of volume mounts for the container (key ['volumeMounts'])
    - env_vars: Environment variables for the container (key ['env'])
    - env_from: List of sources to populate environment variables (key ['envFrom'])
    - image_pull_policy: Policy for pulling the container image (key ['imagePullPolicy'])
    - container_resources: Resource requirements for the container (key ['resources'])
    - container_security_context: Security context for the container (key ['securityContext'])
    - termination_message_policy: Policy for the termination message (key ['terminationMessagePolicy'])
    For a full list, see https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.26/#container-v1-core.
    The following arguments are supported under the pod_metadata argument, which configures the metadata of the pod:
    - name: ['name']
    - namespace: ['namespace']
    - labels: ['labels']
    - annotations: ['annotations']
    For a full list, see https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.26/#objectmeta-v1-meta.

    Example usage:
    .. code-block:: python
        k8s_pod_defs(
            specs=[AssetSpec("my_asset")],
            pod_spec_config={
                "containers": [
                    {
                        "name": "airflow-test-pod",
                        "image": "ubuntu:18.04",
                        "command": ["bash", "-cx"],
                        "args": ["sleep 100"],
                    }
                ]
            },
            pod_metadata={"name": "my-pod"},
            in_cluster=False,
            cluster_context="minikube",
        )

    """

    @multi_asset(specs=specs)
    def _k8s_multi_asset(context: AssetExecutionContext):
        return (
            PipesK8sClient(
                load_incluster_config=load_incluster_config,
                kubeconfig_file=kubeconfig_file,
                kube_context=kube_context,
            )
            .run(
                context=context.op_execution_context,
                base_pod_meta=pod_metadata,
                base_pod_spec=pod_spec,
            )
            .get_results()
        )

    return Definitions(assets=[_k8s_multi_asset])



from collections import defaultdict
from collections.abc import Mapping, Sequence
from functools import cached_property
from typing import AbstractSet, Union  # noqa: UP035

from dagster import AssetKey, AssetsDefinition, AssetSpec
from dagster._annotations import beta, public
from dagster._record import record

from dagster_airlift.core.airflow_instance import AirflowInstance
from dagster_airlift.core.serialization.compute import (
    AirliftMetadataMappingInfo,
    build_airlift_metadata_mapping_info,
)
from dagster_airlift.core.serialization.serialized_data import DagHandle, TaskHandle
from dagster_airlift.core.utils import (
    dag_handles_for_spec,
    is_dag_mapped_asset_spec,
    is_peered_dag_asset_spec,
    is_task_mapped_asset_spec,
    peered_dag_handles_for_spec,
    spec_iterator,
    task_handles_for_spec,
)

MappedAsset = Union[AssetSpec, AssetsDefinition]


@beta
@record
class AirflowDefinitionsData:
    """A class that holds data about the assets that are mapped to Airflow dags and tasks, and
    provides methods for retrieving information about the mappings.
    The user should not instantiate this class directly. It is provided when customizing the events
    that are generated by the Airflow sensor using the `event_transformer_fn` argument of
    :py:func:`build_defs_from_airflow_instance`.
    """

    airflow_instance: AirflowInstance
    airflow_mapped_assets: Sequence[MappedAsset]

    @public
    @property
    def instance_name(self) -> str:
        """The name of the Airflow instance."""
        return self.airflow_instance.name

    @cached_property
    def all_asset_specs(self) -> Sequence[AssetSpec]:
        return list(spec_iterator(self.airflow_mapped_assets))

    @cached_property
    def mapping_info(self) -> AirliftMetadataMappingInfo:
        return build_airlift_metadata_mapping_info(self.airflow_mapped_assets)

    @cached_property
    def all_asset_specs_by_key(self) -> Mapping[AssetKey, AssetSpec]:
        return {spec.key: spec for spec in self.all_asset_specs}

    @public
    def task_ids_in_dag(self, dag_id: str) -> set[str]:
        """Returns the task ids within the given dag_id.

        Args:
            dag_id (str): The dag id.
        """
        return self.mapping_info.task_id_map[dag_id]

    @property
    def dag_ids_with_mapped_asset_keys(self) -> AbstractSet[str]:
        return self.mapping_info.dag_ids

    @cached_property
    def mapped_asset_keys_by_task_handle(self) -> Mapping[TaskHandle, AbstractSet[AssetKey]]:
        asset_keys_per_handle = defaultdict(set)
        for spec in self.all_asset_specs:
            if is_task_mapped_asset_spec(spec):
                task_handles = task_handles_for_spec(spec)
                for task_handle in task_handles:
                    asset_keys_per_handle[task_handle].add(spec.key)
        return asset_keys_per_handle

    @cached_property
    def mapped_asset_keys_by_dag_handle(self) -> Mapping[DagHandle, AbstractSet[AssetKey]]:
        asset_keys_per_handle = defaultdict(set)
        for spec in self.all_asset_specs:
            if is_dag_mapped_asset_spec(spec):
                dag_handles = dag_handles_for_spec(spec)
                for dag_handle in dag_handles:
                    asset_keys_per_handle[dag_handle].add(spec.key)
        return asset_keys_per_handle

    @cached_property
    def peered_dag_asset_keys_by_dag_handle(self) -> Mapping[DagHandle, AbstractSet[AssetKey]]:
        asset_keys_per_handle = defaultdict(set)
        for spec in self.all_asset_specs:
            if is_peered_dag_asset_spec(spec):
                dag_handles = peered_dag_handles_for_spec(spec)
                for dag_handle in dag_handles:
                    asset_keys_per_handle[dag_handle].add(spec.key)
        return asset_keys_per_handle

    @public
    def asset_keys_in_task(self, dag_id: str, task_id: str) -> AbstractSet[AssetKey]:
        """Returns the asset keys that are mapped to the given task.

        Args:
            dag_id (str): The dag id.
            task_id (str): The task id.
        """
        return self.mapped_asset_keys_by_task_handle[TaskHandle(dag_id=dag_id, task_id=task_id)]



from collections.abc import Mapping
from pathlib import Path
from typing import TYPE_CHECKING, Any

import yaml

if TYPE_CHECKING:
    from airflow import DAG


def get_airflow_dags_folder() -> Path:
    # Keep import within fxn to ensure that we perform necessary verification steps first.
    from airflow.configuration import conf

    return Path(conf.get("core", "dags_folder"))


def get_all_dags() -> Mapping[str, "DAG"]:
    from airflow.models import DagBag

    return DagBag().dags


def scaffold_proxied_state(logger: Any) -> None:
    """Scaffolds a proxied state folder for the current Airflow installation.
    Each proxied state is marked as False.
    """
    proxied_state_dir = get_airflow_dags_folder() / "proxied_state"
    if proxied_state_dir.exists():
        raise Exception(
            f"Proxied state directory already exists at {proxied_state_dir}. Please remove this directory before scaffolding."
        )
    logger.info(f"Scaffolding proxied state directory at {proxied_state_dir}")
    for dag_id, dag in get_all_dags().items():
        logger.info(f"Scaffolding proxied state for dag {dag_id}")
        proxied_state_file = proxied_state_dir / f"{dag_id}.yaml"
        proxied_state_file.parent.mkdir(parents=True, exist_ok=True)
        tasks_in_alphabetical_order = sorted(dag.tasks, key=lambda task: task.task_id)
        proxied_state = {
            "tasks": [
                {"id": task.task_id, "proxied": False} for task in tasks_in_alphabetical_order
            ]
        }
        with open(proxied_state_file, "w") as f:
            yaml.dump(proxied_state, f)
